import os
import re
import math
import sys
from argparse import Namespace

from .writers import write_phage_and_bact, write_gff3, write_prophage_tbl
from .writers import write_prophage_tsv, write_genbank, write_prophage_coordinates, write_test_data
from .writers import log_and_message

import PhiSpyRepeatFinder


def find_repeat(fn, st, ppno, extra_dna, output_dir):
    if len(fn) == 0:
        log_and_message("Len sequence is 0 so ignoring\n", c="RED", stderr=True, loglevel="WARNING")
        return {}

    rep = {}
    index = 0

    # with open(os.path.join(output_dir, "repeat_finding"), 'a') as rptout:
    #     rptout.write(f">pp{ppno} {st}\n{fn}\n")

    try:
        # set the False parameter to True to enable debugging of repeat finder
        repeats = PhiSpyRepeatFinder.repeatFinder(fn, 3, ppno, False)
    except Exception as e:
        log_and_message(f"There was an error running repeatfinder for {fn}:{e}\n", c="RED", stderr=True,
                        loglevel="WARNING")
        return {}

    for r in repeats:
        if (r['first_start'] < (3 * extra_dna)) and (r['second_start'] > (len(fn) - (3 * extra_dna))):
            # check that start is always less than end
            # This always causes an off by one error, so we have to increment our ends
            if r['first_end'] < r['first_start']:
                [r['first_start'], r['first_end']] = [r['first_end'] + 1, r['first_start'] + 1]
            if r['second_end'] < r['second_start']:
                [r['second_start'], r['second_end']] = [r['second_end'] + 1, r['second_start'] + 1]

            rep[index] = {}
            rep[index]['s1'] = r['first_start'] + st
            rep[index]['e1'] = r['first_end'] + st
            rep[index]['s2'] = r['second_start'] + st
            rep[index]['e2'] = r['second_end'] + st
            index += 1

    return rep


def check_intg(prophage_sta, prophage_sto, rep, integ, con):
    for m in integ:
        if integ[m]['contig'] != con:
            continue
        if integ[m]['start'] < (prophage_sta + (prophage_sto - prophage_sta) / 2):
            l = integ[m]['start']
        else:
            l = integ[m]['stop']
        if (l - rep['s1'] <= 500 and l - rep['s1'] > 0) or (l - rep['e1'] <= 500 and l - rep['e1'] > 0):
            return 1
        if (l - rep['s2'] <= 500 and l - rep['s2'] > 0) or (l - rep['e2'] <= 500 and l - rep['e2'] > 0):
            return 1
    return 0


def find_smallest(a, b):
    mm = 1000000
    for i in a:
        for j in b:
            if math.fabs(i - j) < mm:
                mm = math.fabs(i - j)
    return mm


def find_rna(prophage_start, prophage_stop, repeat_list, record, cont, integers):
    my_start = 1000000
    start_end = 0
    end_start = 0
    my_end = 1000000
    mydiff = 1000000
    for feature in record.get_entry(cont).get_features('tRNA'):
        i = 0
        while i < len(repeat_list):
            a = find_smallest([feature.start, feature.stop], 
                              [repeat_list[i]['s1'], repeat_list[i]['s2'],
                               repeat_list[i]['e1'], repeat_list[i]['e2']])
            if check_intg(prophage_start, prophage_stop, repeat_list[i], integers, cont) == 1:
                if (math.fabs(repeat_list[i]['s1'] - repeat_list[i]['e2']) > math.fabs(
                        my_start - my_end)) or mydiff == 1000000:
                    my_start = repeat_list[i]['s1']
                    my_end = repeat_list[i]['e2']
                    start_end = repeat_list[i]['e1']
                    end_start = repeat_list[i]['s2']
                    mydiff = a
            if a <= 500 and a < mydiff:
                my_start = repeat_list[i]['s1']
                my_end = repeat_list[i]['e2']
                start_end = repeat_list[i]['e1']
                end_start = repeat_list[i]['s2']
                mydiff = a
            i += 1
    if mydiff == 1000000:
        return '0_0'  # 'null'
    return str(my_start) + '_' + str(my_end)+'_' + str(start_end) + '_' + str(end_start)


def check_pp(contig, start, stop, pp):
    if start > stop:
        (start, stop) = (stop, start)

    for j in pp:
        if contig == pp[j]['contig']:
            if pp[j]['start'] <= start and pp[j]['stop'] >= stop:
                return j
    return 0


def check_phage_word_start(sjcontig, a, b, c):
    j = 0
    tot = 0
    for i in c:
        start = c[i]['start']
        stop = c[i]['stop']
        if start > stop:
            t = start
            start = stop
            stop = t
        if a <= start and stop <= b and c[i]['contig'] == sjcontig:
            if c[i]['pp'] > 0.5:
                j += 1
            tot += 1
    if tot < 4 * j:
        return a
    else:
        return b


def count_phage_word_ends(sjcontig, a, b, c, maxpp):
    j = 0
    tot = 0
    for i in c:
        start = c[i]['start']
        stop = c[i]['stop']
        if start > stop:
            t = start
            start = stop
            stop = t
        if a <= start and stop <= b and c[i]['contig'] == sjcontig:
            if c[i]['pp'] > maxpp:
                j = j + 1
            tot = tot + 1
    return j, tot


def check_phage_word_end(sjcontig, a, b, c):
    j, tot = count_phage_word_ends(sjcontig, a, b, c, 0.5)
    if tot < 4 * j:
        return b
    else:
        return a


def final_check_phage_word(sjcontig, a, b, c):
    j, tot = count_phage_word_ends(sjcontig, a, b, c, 0)
    if j > 5 and tot < 2 * j:
        return str(a) + '_' + str(b)
    else:
        return '0_0'


def clarification_by_phage_word(sjcontig, bef_start, bef_stop, aft_start, aft_stop, genome):
    if aft_start == 0 and aft_stop == 0:
        return '0_0'
    if bef_start <= aft_start:
        s = check_phage_word_start(sjcontig, bef_start, aft_start, genome)
    else:
        s = check_phage_word_start(sjcontig, aft_start, bef_start, genome)

    if bef_stop <= aft_stop:
        e = check_phage_word_end(sjcontig, bef_stop, aft_stop, genome)
    else:
        e = check_phage_word_end(sjcontig, aft_stop, bef_stop, genome)
    se = final_check_phage_word(sjcontig, s, e, genome)
    return se


def fixing_start_end(**kwargs):
    self = Namespace(**kwargs)
    # make all predicted pp list
    log_and_message("Checking prophages we might have found", c="GREEN", stderr=True, quiet=self.quiet)
    pp = {}
    i = 0
    flag = 0
    intg = {}
    intg_index = 1
    genome = {}
    index = 1
    distance_from_last_prophage = 1000
    last_pp = ["" for x in range(9)]

    for this_pp in self.initial_tbl:
        distance_from_last_prophage += 1

        # Find location of all prophage regions
        # columns 7 and 8 are my_status and pp
        if float(this_pp[7]) >= 1 or float(this_pp[8]) >= 1:
            # This is a prophage region. Is it a new prophage?
            new_prophage = False
            # check the sequences are on the same contig. If not, definitely a new prophage
            if this_pp[2] != last_pp[2]:
                new_prophage = True
                distance_from_last_prophage = 1000
            if flag == 0 and distance_from_last_prophage > self.nonprophage_genegaps:
                # we need at least 10 non phage genes betweeen prophages.
                new_prophage = True
            if new_prophage:
                i += 1
                pp[i] = {}
                pp[i]['contig'] = this_pp[2]
                pp[i]['start'] = min(int(this_pp[3]), int(this_pp[4]))
                pp[i]['stop'] = max(int(this_pp[3]), int(this_pp[4]))
                pp[i]['num genes'] = 1
                flag = 1
                last_pp = this_pp
            else:
                pp[i]['stop'] = max(pp[i]['stop'], int(this_pp[3]), int(this_pp[4]))
                pp[i]['num genes'] += 1
            distance_from_last_prophage = 0
        else:
            flag = 0
            # Find location of integrases
        if float(this_pp[8]) == 1.5:
            intg[intg_index] = {}
            intg[intg_index]['start'] = min(int(this_pp[3]), int(this_pp[4]))
            intg[intg_index]['stop'] = max(int(this_pp[3]), int(this_pp[4]))
            intg[intg_index]['contig'] = str(this_pp[2])
            intg_index += 1
        genome[index] = {}
        genome[index]['start'] = int(this_pp[3])
        genome[index]['stop'] = int(this_pp[4])
        genome[index]['pp'] = float(this_pp[8])
        genome[index]['contig'] = this_pp[2]
        genome[index]['function'] = this_pp[1]
        genome[index]['rank'] = float(this_pp[6])
        index += 1


    #######################################################################################
    #                                                                                     #
    # Filter the potential prophages based on how many potential prophage genes are       #
    # in the window, and tell us which ones were dropped.                                 #
    #                                                                                     #
    #######################################################################################
    temppp = {}
    j = 1
    prophagesummary = []
    for i in pp:
        if pp[i]['num genes'] >= self.number:
            temppp[j] = pp[i]
            j += 1
            prophagesummary.append([pp[i]['contig'], pp[i]['start'], pp[i]['stop'], pp[i]['num genes'], "Kept"])
        else:
            prophagesummary.append([pp[i]['contig'], pp[i]['start'], pp[i]['stop'], pp[i]['num genes'],
                                    "Dropped. Not enough genes"])
    # print a list of all prophages and the number of genes
    if prophagesummary:
        log_and_message('Potential prophages (sorted highest to lowest)', stderr=True, quiet=self.quiet)
        log_and_message('Contig\tStart\tStop\tNumber of potential genes\tStatus', stderr=True, quiet=self.quiet)
        for p in sorted(prophagesummary, key=lambda x: x[3], reverse=True):
            log_and_message("\t".join(map(str, p)), stderr=True, quiet=self.quiet)
    pp = temppp
    # End filtering
    # find start end for all pp using repeat finder
    dna = {entry.id: str(entry.seq) for entry in self.record}
    extra_dna = 2000
    for i in pp:
        log_and_message(f"PROPHAGE: {i} Contig: {pp[i]['contig']} Start: {pp[i]['start']} Stop: {pp[i]['stop']}",
                        c="PINK", stderr=True, quiet=self.quiet)
        start = pp[i]['start'] - extra_dna
        if start < 1:
            start = 1
        if 'stop' in pp[i]:
            stop = pp[i]['stop'] + extra_dna
        else:
            stop = genome[len(genome) - 1]['stop']
        if stop > len(dna[pp[i]['contig']]):
            stop = len(dna[pp[i]['contig']])
        if stop - start > 200000:
            log_and_message(f"Not checking repeats for pp {i} because it is too big: {stop - start} bp",
                            c="PINK", stderr=True, quiet=self.quiet)
            continue
        repeat_list = find_repeat(dna[pp[i]['contig']][start:stop], start, i, extra_dna, self.output_dir)
        s_e = find_rna(start, stop, repeat_list, self.record, pp[i]['contig'], intg)
        if s_e != 'null':
            t = re.split('_', s_e)
            if s_e == '0_0':
                s_e1 = clarification_by_phage_word(pp[i]['contig'], pp[i]['start'], pp[i]['stop'], pp[i]['start'],
                                                   pp[i]['stop'], genome)
            else:
                s_e1 = clarification_by_phage_word(pp[i]['contig'], pp[i]['start'], pp[i]['stop'], float(t[0]),
                                                   float(t[1]), genome)
            t1 = re.split('_', s_e1)
            # we only accept this if it increases the phage length
            if float(t1[0]) > 0 and float(t1[0]) < pp[i]['start'] and float(t1[1]) > 0 and float(t1[1]) > pp[i]['stop']:
                pp[i]['start'] = float(t1[0])
                pp[i]['stop'] = float(t1[1])

            if (float(t[0]) != 0) and (pp[i]['start'] == float(t[0])) and (pp[i]['stop'] == float(t[1])):
                # 1 if (float(t[0])!= 0) and (float(t1[0]) == float(t[0])) and (float(t1[1]) == float(t[1])):
                temps1 = min(t[0], t[2])
                tempe1 = max(t[0], t[2])
                temps2 = min(t[1], t[3])
                tempe2 = max(t[1], t[3])
                pp[i]['att'] = [
                    t[0],
                    t[2], 
                    t[3],
                    t[1], 
                    dna[pp[i]['contig']][int(temps1) - 1:int(tempe1)],
                    dna[pp[i]['contig']][int(temps2) - 1:int(tempe2)],
                    'Repeat exactly at the end'
                ]
                # string representation of the above
                pp[i]['atts'] = "\t".join(map(str, pp[i]['att']))
            else:
                # this approach will just append the longest repeat
                longestrep = 0
                bestrep = None
                samelenrep = 0
                for idx in repeat_list:
                    lengthrep = math.fabs(repeat_list[idx]['e1'] - repeat_list[idx]['s1'])
                    if lengthrep > longestrep and lengthrep < 150:
                        longestrep = lengthrep
                        bestrep = repeat_list[idx]
                        samelenrep = 1
                    elif lengthrep == longestrep:
                        samelenrep += 1
                if bestrep:
                    attLseq = dna[pp[i]['contig']][int(bestrep['s1']) - 1:int(bestrep['e1']) - 1]
                    attRseq = dna[pp[i]['contig']][int(bestrep['s2']) - 1:int(bestrep['e2']) - 1]
                    if len(attLseq) == 0:
                        msg = f"The attL sequence had no length from {int(bestrep['s1']) - 1} to "
                        msg += f"{int(bestrep['e1']) - 1} on contig {pp[i]['contig']} "
                        msg += f"(length: {len(dna[pp[i]['contig']])} )"
                        log_and_message(msg, c="YELLOW", stderr=True, quiet=self.quiet)
                    if len(attRseq) == 0:
                        msg = f"The attL sequence had no length from {int(bestrep['s2']) - 1} to "
                        msg += f"{int(bestrep['e2']) - 1} on contig {pp[i]['contig']} "
                        msg += f"(length: {len(dna[pp[i]['contig']])} )"
                        log_and_message(msg, c="YELLOW", stderr=True, quiet=self.quiet)
                    pp[i]['att'] = [
                        bestrep['s1'],
                        bestrep['e1'],
                        bestrep['s2'],
                        bestrep['e2'],
                        attLseq,
                        attRseq,
                        "Longest Repeat flanking phage and within " + str(extra_dna) + " bp"
                    ]
                    pp[i]['atts'] = "\t".join(map(str, pp[i]['att']))
                    if samelenrep > 1:
                        msg=f"There were {samelenrep} repeats with the same length as the best. One chosen somewhat randomly!"
                        log_and_message(msg, c="YELLOW", stderr=True, quiet=self.quiet)
    # fix start end for all pp
    try:
        outfile = open(os.path.join(self.output_dir, self.file_prefix + 'prophage_information.tsv'), 'w')
    except IOError as e:
        log_and_message(f"There was an error opening the files: {e}\n", c="RED", stderr=True, loglevel="CRITICAL")
        sys.exit(-1)

    outfile.write("identifier\tfunction\tcontig\tstart\tstop\tposition\trank\tmy_status\tpp\tFinal_status\t")
    outfile.write("start of attL\tend of attL\tstart of attR\tend of attR\tsequence of attL\tsequence of attR\t")
    outfile.write("Reason for att site\n")

    for this_pp in self.initial_tbl:
        me = check_pp(this_pp[2], int(this_pp[3]), int(this_pp[4]), pp)
        if me == 0:
            outfile.write("\t".join(map(str, this_pp + [0])) + '\n')
        else:
            outfile.write("\t".join(map(str, this_pp)) + '\t' + str(me) + '\n')

    """
    now we need to decide which files to keep
    It is based on this code:
        Code | File
        --- | ---
        1 | prophage_coordinates.tsv 
        2 | GenBank format output 
        4 | prophage and bacterial sequences  
        8 | prophage_information.tsv  
        16 | prophage.tsv  
        32 | GFF3 format  
        64 | prophage.tbl
        128 | test data used in the random forest
    As explained in the README. 
    """

    oc = self.output_choice

    if oc >= 128:
        # write the calculated data
        write_test_data(self, pp)
        oc -= 128
    if oc >= 64:
        # write the prophage location table
        write_prophage_tbl(self, pp)
        oc -= 64
    if oc >= 32:
        # write the prophage in GFF3 format
        write_gff3(self, pp)
        oc -= 32
    if oc >= 16:
        # write a tsv file of this data
        write_prophage_tsv(self, pp)
        oc -= 16
    if oc < 8:
        # we want to KEEP the prophage_information.tsv that we created earlier
        # if oc >= 8, or if we get to this point and oc is 7 or less we want to
        # delete it.
        os.remove(os.path.join(self.output_dir, self.file_prefix + 'prophage_information.tsv'))
    if oc == 8:
        oc = 0
    if oc >= 4:
        # separate out the bacteria and phage as fasta files
        write_phage_and_bact(self, pp, dna)
        oc -= 4
    if oc >= 2:
        # update input GenBank file and incorporate prophage regions
        write_genbank(self, pp)
    if oc >= 1:
        # print the prophage coordinates:
        write_prophage_coordinates(self, pp)









"""
Test suite for seqio_filter.py

This is (currently) not complete, and is a work in progres.
"""

import unittest
import gzip

from Bio import SeqIO, SeqFeature
from Bio.SeqFeature import FeatureLocation
from PhiSpyModules import SeqioFilter


__license__ = 'MIT'
__maintainer__ = 'Rob Edwards'
__email__ = 'raedwards@gmail.com'





class SeqIO_FilterTest(unittest.TestCase):
    def test_merge_split(self):
        """
        Test the merge and split function

        Paracoccus_yeei_TT13.gb has the following compound locations:
        PYTT13_06780: join{[1366920:1367182](-), [1365992:1366921](-)}
                      [1365992:1367182](-)

        PYTT13_11395: join{[2283890:2284152](+), [2284151:2285080](+)}
                      [2283890:2285080](+)

        PYTT13_11465: join{[2301567:2301817](+), [2301816:2302655](+)}
                      [2301567:2302655](+)

        PYTT13_12460: join{[2495319:2495581](+), [2495580:2496509](+)}
                      [2495319:2496509](+)

        PYTT13_16505: join{[3331106:3331356](+), [3331355:3332194](+)}
                      [3331106:3332194](+)

        :return:
        """
        correct_locations = {
            'PYTT13_06780': (1365992, 1367182),
            'PYTT13_11395': (2283890, 2285080),
            'PYTT13_11465': (2301567, 2302655),
            'PYTT13_12460': (2495319, 2496509),
            'PYTT13_16505': (3331106, 3332194)
        }

        testgbk = "test_genbank_files/Paracoccus_yeei_TT13.gb.gz"
        handle = gzip.open(testgbk, 'rt')
        record = SeqioFilter(SeqIO.parse(handle, "genbank"))
        handle.close()
        for s in record:
            for f in s.get_features("CDS"):
                if 'locus_tag' in f.qualifiers and f.qualifiers['locus_tag'][0] in correct_locations:
                    lt = f.qualifiers['locus_tag'][0]
                    self.assertIsInstance(f.location,FeatureLocation)
                    self.assertEqual(f.location.start, correct_locations[lt][0])
                    self.assertEqual(f.location.end, correct_locations[lt][1])


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
import os
import sys
import argparse
import re
import pkg_resources
import binascii
import logging

from .formatting import message

try:
    __version__ = pkg_resources.get_distribution('phispy').version
except Exception:
    __version__ = 'unknown'

def print_list():
    f = None
    try:
        # with pip we use resource streams that may be files or from archives
        f = pkg_resources.resource_stream('PhiSpyModules', 'data/trainingGenome_list.txt')
    except:
        message('Cannot find the list of training sets. It should be in data/trainingGenome_list.txt', "RED", 'stderr')
        sys.exit(-1)
    for line in f:
        line = line.decode().strip()
        temp = re.split('\t', line)
        if int(temp[3]) == 1:
            print("{}\t{}".format(temp[2], 'data/' + temp[1]))
    f.close()


def is_valid_file(x):
    if not x or not os.path.exists(x):
        raise argparse.ArgumentTypeError("Checking for validity: {0} does not exist".format(x))
    return x


def is_gzip_file(f):
    """
    This is an elegant solution to test whether a file is gzipped by reading the first two characters.
    I also use a version of this in fastq_pair if you want a C version :)

    See https://stackoverflow.com/questions/3703276/how-to-tell-if-a-file-is-gzip-compressed for inspiration
    :param f: the file to test
    :return: True if the file is gzip compressed else false
    """
    with open(f, 'rb') as i:
        return binascii.hexlify(i.read(2)) == b'1f8b'

def create_logger(self):
    """
    Add a logger to self
    :param self: the args parsed object
    :return: a logger for the args parsed object
    """
    if not self.log:
        self.log = os.path.join(self.output_dir, self.file_prefix + 'phispy.log')

    logger = logging.getLogger('PhiSpy')
    logger.setLevel(5)
    hdlr = logging.FileHandler(self.log)
    fmt = logging.Formatter('%(asctime)s %(levelname)-8s %(message)s', datefmt='%Y-%m-%d %H:%M:%S')
    hdlr.setFormatter(fmt)
    logger.addHandler(hdlr)
    return logger

def get_args():
    parser = argparse.ArgumentParser(
        description="phiSpy is a program for identifying prophages from among microbial genome sequences",
        epilog="(c) 2008-2018 Sajia Akhter, Katelyn McNair, Przemysław Decewicz, Rob Edwards, " +
               "San Diego State University, San Diego, CA")
    parser.add_argument('infile', type=is_valid_file, help='Input file in genbank format', nargs='?')
    parser.add_argument('-m', '--make_training_data', type=str,
                        help='Create training data from a set of annotated genome files. Requires is_phage=1 ' +
                             'qualifier in prophage\'s CDSs')
    parser.add_argument('-t', '--training_set', action='store', default='data/trainSet_genericAll.txt',
                        help='Choose the most closely related set to your genome.')
    parser.add_argument('-l', '--list', action='store_true', default=False,
                        help='List the available training sets and exit')
    parser.add_argument('-p', '--file_prefix', default="",
                        help='An optional prefix to prepend to all of the output files')
    parser.add_argument('-e', '--evaluate', type=bool, default=False, const=True, nargs='?',
                        help='Run in evaluation mode -- does not generate new data, but reruns the evaluation')
    parser.add_argument('-n', '--number', default=5, type=int,
                        help='Number of consecutive genes in a region of window size that must be prophage genes' +
                             ' to be called. [Default: 5]')
    parser.add_argument('-u', '--min_contig_size', default=5000, type=int,
                        help='Minimum contig size (in bp) to be included in the analysis. Smaller contigs will ' +
                             'be dropped. [Default: 5000]')
    parser.add_argument('-w', '--window_size', default=30, type=int,
                        help='Window size of consecutive genes to look through to find phages. [Default: 30]')
    parser.add_argument('-g', '--nonprophage_genegaps', default=10, type=int,
                        help='The number of non phage genes betweeen prophages. [Default: 10]')
    parser.add_argument('-r', '--randomforest_trees', default=500, type=int,
                        help='Number of trees generated by Random Forest classifier. [Default: 500]')
    parser.add_argument('--expand_slope', action='store_true', default=False,
                        help='Use the product of the slope of the Shannon scores in making test sets')
    parser.add_argument('--kmers_type', default='all', choices=['all', 'codon', 'simple'], type=str,
                        help='Type of kmers used for calculating Shannon scores. [Default: all]')
    parser.add_argument('--phmms', type=str,
                        help='Phage HMM profile database (like pVOGs) will be mapped against the genome of ' +
                             'interest and used as additional feature to identify prophages.')
    parser.add_argument('--skip_search', action='store_true',
                        help='If set, the search part will be skipped and the program will assume the ' +
                             'existance of updated GenBank files.')
    parser.add_argument('--color', action='store_true',
                        help='If set, within the output GenBank file CDSs with phmms hits will be ' +
                             'colored (for viewing in Artemis).')
    parser.add_argument('--threads', type=int, default=2,
                        help='Number of threads to use while searching with phmms and the random forest.')
    parser.add_argument('-o', '--output_dir', help='The output directory to write the results')
    parser.add_argument('--output_choice', type=int, default=3,
                        help='Sum of codes for files to output. For more details see the README.md file')
    parser.add_argument('--log', type=str,
                        help="Name of the log file to write details to (defulat: phispy.log)")
    parser.add_argument('--quiet', action='store_true',
                        help='Run in quiet mode')
    parser.add_argument('-k', '--keep', type=bool, default=False, const=True, nargs='?',
                        help='Do not delete temp files')
    parser.add_argument('-v', '--version', action='version', version=__version__)
    args = parser.parse_args()


    ######################################
    #   list the training sets and exit  #
    ######################################
    if args.list:
        print_list()
        exit(0)

    if args.file_prefix != "":
        if not args.file_prefix.endswith("_"):
            args.file_prefix += "_"
        args.file_prefix = args.file_prefix.replace(' ', '_')

    # check whether output directory was provided
    if not args.output_dir and not args.make_training_data:
        message("ERROR: Output directory (-o) is required. Use -h for more options\n", "RED", "stderr")
        sys.exit(-1)
    elif args.output_dir:
        os.makedirs(args.output_dir, exist_ok=True)
    else:
        args.output_dir=""

    args.logger = create_logger(args)
    return args




import types
from Bio import SeqFeature
import copy

from .writers import log_and_message
from Bio.SeqFeature import FeatureLocation, CompoundLocation

class SeqioFilter( list ):
    """This class is to allow filtering of the Biopython SeqIO record

    SeqIO.parse returns a generator object so anytime you want to perform
    an action on it, you must iterate through the entire list. This
    class adds the ability to filter and return only a subset of the
    features. Those are split into separate features or joined based
    on the distance between the parts and then sorted based on the
    start to end location.

    Note:
        To use simply pass a SeqIO.parse object to it and then when
        the object is called a keyword is passed to it and only those
        features matching the keyword are returned.
    Example:
        record = SeqioFilter(SeqIO.parse(infile))):
        #no change to standard SeqIO calls
        for entry in record:
            print(entry.id, entry.seq)
        #now we can get only certain features
        for cds in record.get_features('CDS'):
            print(cds)

    """

    def __init__( self, content ):
        self.n = 0
        self.get_n = dict()
        for n, item in enumerate(content):
            self.attach_methods(item)
            self.get_n[item.id] = n
            self.append(item)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = self[self.n]
        except IndexError:
            self.n = 0
            raise StopIteration()
        self.n += 1
        return item

    def __call__(self, keyword=''):
        pass

    def get_entry(self, id):
        return self[self.get_n[id]]

    def merge_or_split(self, seq, feature, mindistance = 100):
        """
        Merge or split sequence features with compound locations
        :param seq: the sequence object
        :param feature: The feature with a compound location to merge or split
        :param mindistance: the distance with which they will be merged/split
        :return:
        """
        thisid = " ".join(feature.qualifiers.get('locus_tag', [str(feature.location)]))
        # find the index of this in the list. We could also pass this as an arg
        idx = seq.features.index(feature)

        # do we need to do anything
        if type(feature.location) != CompoundLocation:
            log_and_message(f"Error {thisid} does not appear to be a compound location\n",
                            c="RED", stderr=True, loglevel="WARNING")
            return

        log_and_message(f"merging/splitting {thisid} original location: {feature.location}")

        if 'product' in feature.qualifiers:
            feature.qualifiers['product'][0] = 'Merged-or-split: ' + feature.qualifiers['product'][0]
        else:
            feature.qualifiers['product'] = ['Merged-or-split: not assigned']


        # simplify our coding
        loc = feature.location
        # first we find the strand
        strand = loc.parts[0].strand

        # test that all locations are on the same strand
        # record an error if not
        for p in loc.parts:
            if p.strand != strand:
                msg = "Error: We can not handle compound locations on different strands. For {thisid} we have {loc}"
                log_and_message(msg, c="RED", stderr=True, loglevel="WARNING")
                return

        all_locs = []
        merged = loc.parts[0]
        # handle features on the + strand
        if strand > 0:
            for p in loc.parts[1:]:
                if p.start < merged.start:
                    # this feature spans a break
                    msg = (f"Feature {thisid} spans the origin: {loc}\n"
                           f"WARNING: THIS IS AN UNTESTED FEATURE!\n"
                           f"We have not thoroughly tested conditions where an ORF on the +ve strand appears to cross "
                           f"the origin of the contig. We would appreciate you posting an issue on GitHub and sending "
                           f"Rob a copy of your genome to test!\n")
                    log_and_message(msg, c="YELLOW", stderr=True, loglevel='WARNING')

                    all_locs.append(merged)
                    merged = p
                    continue
                if p.start > merged.start and p.start < merged.end:
                    merged = FeatureLocation(merged.start, p.end, strand)
                elif p.start > merged.end:
                    if p.start - merged.end > mindistance:
                        all_locs.append(merged)
                        merged = FeatureLocation(p.start, p.end, strand)
                    else:
                        merged = FeatureLocation(merged.start, p.end, strand)
                else:
                    all_locs.append(merged)
                    merged = FeatureLocation(p.start, p.end, strand)
        # handle features on the -ve strand
        else:
            for p in loc.parts[1:]:
                if merged.start < p.start:
                    # this feature spans a break
                    msg = (f"Feature {thisid} spans the origin: {loc}\n"
                           f"WARNING: THIS IS AN UNTESTED FEATURE!\n"
                           f"We have not thoroughly tested conditions where an ORF on the -ve strand appears to cross "
                           f"the origin of the contig. We would appreciate you posting an issue on GitHub and sending "
                           f"Rob a copy of your genome to test!\n")
                    log_and_message(msg, c="YELLOW", stderr=True, loglevel='WARNING')

                    all_locs.append(merged)
                    merged = p
                    continue
                if p.end > merged.start and p.end < merged.end:
                    # trivial case, the ORFs overlap
                    merged = FeatureLocation(p.start, merged.end, strand)
                elif p.end < merged.start:
                    # more complex case, there is a gap
                    if merged.start - p.end > mindistance:
                        all_locs.append(merged)
                        merged = FeatureLocation(p.start, p.end, strand)
                    else:
                        merged = FeatureLocation(p.start, merged.end, strand)
                else:
                    all_locs.append(merged)
                    merged = FeatureLocation(p.start, p.end, strand)

        if all_locs:
            # we have multiple features, so we need to add features
            # make sure we add the last feature
            all_locs.append(merged)
            log_and_message(f"We could not join the whole {feature.type} feature into a single new feature.")
            # replace the existing compound feature with the first one of the split features
            newfeat = feature
            newfeat.location = all_locs[0]
            seq.features[idx] = newfeat
            # append the other features
            for f in all_locs[1:]:
                newfeat = copy.deepcopy(feature)
                newfeat.location = f
                seq.features.append(newfeat)
                log_and_message(f"Appended part of a multiple {newfeat.type} feature {thisid} loc: {f}\n")
        else:
            # we just replace the old feature with the new
            # update the location
            feature.location = merged
            # add the new feature
            seq.features[idx] = feature
            log_and_message(f"Created a single {feature.type} feature: {thisid} loc: {merged}\n")

    def attach_methods(self, target):
        """This method allows attaching new methods to the SeqIO entry object

           Args:
               target: is the SeqIO object that will be attaching a method to
        """

        # resolve compound locations!
        for feat in target.features:
            if type(feat.location) == CompoundLocation:
                self.merge_or_split(target, feat)

        # sort the features based on their location in the genome
        target.features.sort( key = lambda feature : tuple([min(feature.location.start, feature.location.end),
                                                            max(feature.location.start, feature.location.end)]) )

        def get_features(target, feature_type):
            for feature in target.features:
                feature.id       = " ".join(feature.qualifiers.get('locus_tag', [str(feature.location)]))
                feature.function = " ".join(feature.qualifiers.get('product',['unknown']))
                # set start and stop for the feature and swap for direction if needed
                feature.start    = int(feature.location.start) + 1
                feature.stop     = int(feature.location.end)
                feature.phmm     = [1.0] if 'phmm' not in feature.qualifiers else [float(x.split(':')[1]) for x in feature.qualifiers.get('phmm')]
                if feature.strand < 0:
                    feature.start, feature.stop = feature.stop, feature.start
                # if a feature type was provided, only return those features
                if not feature_type or feature.type == feature_type:
                    yield feature
        target.get_features = types.MethodType(get_features,target)

        def append_feature(target, feature):
            target.features.append(feature)

        target.append_feature = types.MethodType(append_feature, target)



#!/usr/bin/env python
"""
Create a prophage.tbl file from a phispy directory that does not contain one.
"""

import os
import sys
import argparse

INSTALLATION_DIR = os.path.dirname(os.path.dirname(__file__))
sys.path.append(INSTALLATION_DIR)

from .writers import  prophage_measurements_to_tbl

def main(phispydir):
    """
    Make a new prophage table
    :param phispydir: the directory to read the input and create the output
    :return: nothing
    """
    prophage_measurements_to_tbl(os.path.join(phispydir, 'prophage_tbl.txt'), os.path.join(phispydir, 'prophage.tbl'))

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="create a prophage.tbl file for a PhiSpy directory")
    parser.add_argument('-d', help='phispy directory')
    parser.add_argument('-v', help='verbose output', action="store_true")
    args = parser.parse_args()
    main(args.d)



"""
A custom error class for PhiSpy to make the code more accessible
"""
class Error(Exception):
    """
    Base class for exceptions in this module.
    """
    pass

class ColorNotFoundError(Error):
    """
    Exception raised for sequences not being paired properly.

    :param message: explanation of the error
    """

    def __init__(self, message):
        self.message = message
        super().__init__(self.message)

class NoBasesCounted(Error):
    """
    An exception raised when no AT or GC bases are counted
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)





# ==============
# Type classes
# PathType
# This argparse Type allows for verification the passed argument is a path, but
# not opening a file immediately.
#
# This can be useful for multiple things.
# - 1) In case of an unrelated error causes the script to crash or hang
#      the file will not be left open, as it was never opened.
#      exist without having ever created an output file.
# - 2) Can be used to specify
#
# ==============
# Example Usage
# Check for a valid and existing file.

# parser = argparse.ArgumentParser()
#
# # Add an argument that must be an existing file, but can also be specified as a dash ('-') in the command.
# parser.add_argument('existing_file', type = PathType(type='file', dash_ok=True, exists = True))
#
# # Add an argument for a folder that must NOT exist. Note: folders can not be "dash_ok".
# parser.add_argument('non_existant_folder', type = PathType(type='dir', exists = False))
#
# # Add an argument for EITHER a folder or a file, but can be dash_ok, and don't check if it exists.
# parser.add_argument('maybe_existant_file_or_folder',type=PathType(type=('dir','file').__contains__,dash_ok=True,exists=None))

# ==============

# Code taken from https://stackoverflow.com/questions/11415570/directory-path-types-with-argparse
# This is EXACTLY what I was going to build otherwise.
#
# This is also suggested to be an official change to argparse.
# https://mail.python.org/pipermail/stdlib-sig/2015-July/000990.html

import os
from argparse import ArgumentTypeError

# Existance Options
EXIST_CHECK  = True  # File/folder must exist.
EXIST_INVERT = False # Must *not* exist.
EXIST_EITHER = None  # Allowed to reguardless.

# Types allowed for files.
TYPE_FILE = 'file'
TYPE_DIR  = 'dir'
TYPE_SYM  = 'symlink'
TYPE_ALL  = None # Any thing is fine (don't check)

# Whether folder is not empty
EMPTY_CHECK  = True
EMPTY_INVERT = False
EMPTY_EITHER = None

class PathType(object):
    def __init__(self, exists = EXIST_CHECK, type = TYPE_FILE, empty = EMPTY_CHECK, dash_ok = True):
        """
           exists:
                True: a path that does exist
                False: a path that does not exist, in a valid parent directory
                None: don't care
           type: 'file', 'dir', 'symlink', None, a list of these,
                 or a function returning True for valid paths
                 If None, the type of file/directory/symlink is not checked.
           dash_ok: whether to allow "-" as stdin/stdout
        """
        assert exists in (EXIST_CHECK, EXIST_INVERT, EXIST_EITHER)
        assert empty  in (EMPTY_CHECK, EMPTY_INVERT, EMPTY_EITHER)
        # Make sure type is file, dir, sym, None, list, or callable.
        if isinstance(type, (list, tuple)):
            # Type is a list, make sure that it includes only "file", "dir", or "sym"
            for t in type:
                assert t in (TYPE_FILE, TYPE_DIR, TYPE_SYM)
            # Type is the contains check for this lists.
            type = type.__contains__
        elif not callable(type):
            # Otherwise, make sure this is valid.
            assert type in (TYPE_FILE, TYPE_DIR, TYPE_SYM, EXIST_EITHER)
        # else; type is a callable object, and this is ok.
        self._exists = exists
        self._type = type
        self._empty = empty
        self._dash_ok = dash_ok

    def __call__(self, string):
        if string == '-':
            # the special argument "-" means sys.std[in/out]
            if self._type == TYPE_DIR:
                raise ArgumentTypeError('standard input/output (-) not allowed as directory path')
            elif self._type == TYPE_SYM:
                raise ArgumentTypeError('standard input/output (-) not allowed as symlink path')
            elif not self._dash_ok:
                raise ArgumentTypeError('standard input/output (-) not allowed')
            return string # No reason to check anything else if this works.
        # If the file must exist.
        if self._exists == EXIST_CHECK:
            if not os.path.exists(string):
                raise ArgumentTypeError("path does not exist: %r" % string)
            # If must be file and not: Fail
            if self._type == TYPE_FILE:
                if not os.path.isfile(string):
                    raise ArgumentTypeError("path is not a file: %r" % string)
            # If must be dir and not: Fail
            elif self._type == TYPE_DIR:
                if not os.path.isdir(string):
                    raise ArgumentTypeError("path is not a directory: %r" % string)
            # If must be sym and not: Fail
            elif self._type == TYPE_SYM:
                if not os.path.islink(string):
                    raise ArgumentTypeError("path is not a symlink: %r" % string)
            # Otherwise, call type.
            elif not self._type(string):
                raise ArgumentTypeError("path not valid: %r" % string)
        else:
            if self._exists == False and os.path.exists(string):
                raise ArgumentTypeError("path exists: %r" % string)
            p = os.path.dirname(string) or os.path.curdir
            if not os.path.exists(p):
                raise ArgumentTypeError("parent directory does not exist: %r" % p)
            elif not os.path.isdir(p):
                raise ArgumentTypeError("parent path is not a directory: %r" % p)
        # If the file must be empty
        if self._empty == EMPTY_CHECK:
            if os.path.exists(string) and os.listdir(string):
                raise ArgumentTypeError("folder is not empty: %r" % string)
        return string



import os
import re
import math
import sys
import pkg_resources
from argparse import Namespace

from .formatting import message
from .errors import NoBasesCounted

class ShannonScore:
    def __init__(self, kmers_type):
        # Create a hash of the kmers that points to the index of an array that holds the value
        self._kmers = {}
        self._kmers_phage = []
        self._kmers_all = []
        self._kmers_type = kmers_type
        kmers_file = 'data/phage_kmers_' + self._kmers_type + '_wohost.txt'
        if not pkg_resources.resource_exists:
            sys.exit("ERROR: Kmers file {} not found".format(kmers_file))

        for line in pkg_resources.resource_stream('PhiSpyModules', kmers_file):
            line = line.decode().strip()
            self._kmers[line] = ''

    def reset(self):
        self._kmers_phage = []
        self._kmers_all = []

    def addValue(self, seq):
        mer = 12
        pos = 0
        self._kmers_phage.append([])
        self._kmers_all.append(0)
        kmers = self.kmerize_orf(seq, mer, self._kmers_type)
        for kmer in kmers:
            self._kmers_all[-1] += 1
            try:            
                self._kmers[kmer]
                self._kmers_phage[-1].append(kmer)
            except KeyError:
                continue

    def getSlope(self, start, stop):
        total = sum(self._kmers_all[start : stop])
        found_total = sum([len(x) for x in self._kmers_phage[start : stop]])
        if total == 0:
            return 0
        H = 0.0
        window_kmers = {}
        for kl in self._kmers_phage[start : stop]:
            for k in kl:
                try:
                    window_kmers[k] += 1.0
                except KeyError:
                    window_kmers[k] = 1.0
        for i in window_kmers.values():
            p = i/total
            H = H + p * (math.log(p)/math.log(2))
        if H >= 0:
            return 0
        freq_found = found_total / float(total)
        myslope = - freq_found / H
        return myslope

    def kmerize_orf(self, orf, k, t):

        kmers = []
        if t == 'simple':
            stop = len(orf) - (len(orf) % k)
            for i in range(0, stop, k):
                kmers.append(orf[i : i + k])
        elif t == 'all':
            for j in range(0, k):
                stop = len(orf) - ((len(orf) - j) % k)
                for i in range(j, stop, k):
                    kmers.append(orf[i : i + k])
        elif t == 'codon':
            for j in range(0, k, 3):
                stop = len(orf) - ((len(orf) - j) % k)
                for i in range(j, stop, k):
                    kmers.append(orf[i : i + k])

        return kmers

def my_sort(orf_list):
    n = len(orf_list)
    i = 0
    while i < n:
        j = i + 1
        while j < n:
            flag = 0
            # direction for both
            if orf_list[i]['start'] < orf_list[i]['stop']:
                dir_i = 1
            else:
                dir_i = -1
            if orf_list[j]['start'] < orf_list[j]['stop']:
                dir_j = 1
            else:
                dir_j = -1
            # check whether swap need or not
            if dir_i == dir_j:
                if orf_list[i]['start'] > orf_list[j]['start']:
                    flag = 1
            else:
                if dir_i == 1:
                    if orf_list[i]['start'] > orf_list[j]['stop']:
                        flag = 1
                else:
                    if orf_list[i]['stop'] > orf_list[j]['start']:
                        flag = 1
            # swap
            if flag == 1:
                temp = orf_list[i]
                orf_list[i] = orf_list[j]
                orf_list[j] = temp
            j += 1
        i += 1
    return orf_list

def find_all_median(orf_list):
    all_len = []
    for i in orf_list:
        all_len.append((abs(i['start'] - i['stop'])) + 1)
    return find_median(all_len)

def find_median(all_len):
    n = int(round(len(all_len) / 2))
    all_len.sort()
    if len(all_len) == n * 2:
        return (all_len[n] + all_len[n - 1]) / float(2)
    else:
        return all_len[n]

def find_atgc_skew(seq):
    total_at = 0.0
    total_gc = 0.0
    a = 0
    t = 0
    c = 0
    g = 0
    for base in seq:
        if base == 'A':
            a += 1
            total_at += 1
        elif base == 'T':
            t += 1
            total_at += 1
        elif base == 'G':
            g += 1
            total_gc += 1
        elif base == 'C':
            c += 1
            total_gc += 1
        elif base == 'R':
            a += 0.5
            total_at += 0.5
            g += 0.5
            total_gc += 0.5
        elif base == 'Y':
            c += 0.5
            total_gc += 0.5
            t += 0.5
            total_at += 0.5
        elif base == 'S':
            g += 0.5
            total_gc += 0.5
            c += 0.5
            total_gc += 0.5
        elif base == 'W':
            a += 0.5
            total_at += 0.5
            t += 0.5
            total_at += 0.5
        elif base == 'K':
            g += 0.5
            total_gc += 0.5
            t += 0.5
            total_at += 0.5
        elif base == 'M':
            c += 0.5
            total_gc += 0.5
            a += 0.5
            total_at += 0.5
        elif base == 'B':
            c += 0.3
            total_gc += 0.3
            g += 0.3
            total_gc += 0.3
            t += 0.3
            total_at += 0.3
        elif base == 'D':
            a += 0.3
            total_at += 0.3
            g += 0.3
            total_gc += 0.3
            t += 0.3
            total_at += 0.3
        elif base == 'H':
            a += 0.3
            total_at += 0.3
            c += 0.3
            total_gc += 0.3
            t += 0.3
            total_at += 0.3
        elif base == 'V':
            a += 0.3
            total_at += 0.3
            c += 0.3
            total_gc += 0.3
            g += 0.3
            total_gc += 0.3
        elif base == 'N':
            a += 0.25
            total_at += 0.25
            t += 0.25
            total_at += 0.25
            g += 0.25
            total_gc += 0.25
            c += 0.25
            total_gc += 0.25
        else:
            print("seq", seq)
            print("base", base)
            sys.exit("ERROR: Non nucleotide base found")
    if(total_at * total_gc) == 0:
        raise NoBasesCounted("a total of zero total_at*total_gc")
    return float(a)/total_at, float(t)/total_at, float(g)/total_gc, float(c)/total_gc

def find_avg_atgc_skew(orf_list, mycontig, dna):
    a_skew = []
    t_skew = []
    g_skew = []
    c_skew = []
    for i in orf_list:
        start = i['start']
        stop = i['stop']
        try:
            if start < stop:
                bact = dna[mycontig][start - 1:stop]
                xa, xt, xg, xc = find_atgc_skew(bact)
            else:
                bact = dna[mycontig][stop - 1:start]
                xt, xa, xc, xg = find_atgc_skew(bact)
        except NoBasesCounted as e:
            sys.stderr.write(e.message + "\n")
            sys.stderr.write(f"No bases were counted for orf {i} from {start} to {stop}\n")
            sys.stderr.write("This error is usually thrown with an exceptionally short ORF that is only a " +
                             " few bases. You should check this ORF and confirm it is real!\n")
            sys.exit()
        if len(bact) < 3:
            continue
        a_skew.append(xa)
        t_skew.append(xt)
        g_skew.append(xg)
        c_skew.append(xc)
    return a_skew, t_skew, g_skew, c_skew

def reverse_complement(seq):

    rcd = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C', 
           'R': 'Y', 'Y': 'R', 'M': 'K', 'K': 'M',
           'S': 'S', 'W': 'W',
           'H': 'D', 'D': 'H',
           'B': 'V', 'V': 'B',
           'N': 'N'}
    rcseq = ''
    for base in seq[::-1]:
        try:
            rcseq += rcd[base]
        except KeyError:
            rcseq += base

    return rcseq

def measure_features(**kwargs):
    """
    This is a combination of the old make_test_set // make_train_set
    but now returns an array of the data and so reduces redundancy
    :param kwargs: The keyworded arguments object
    :return: an array of arrays. Each element contains median orf length, shannon slope,
            at_skew, gc_skew, max_direction, phmms
    """

    self = Namespace(**kwargs)
    data = []  # the array of arrays we will return
    my_shannon_scores = ShannonScore(self.kmers_type)
    all_orf_list = {}
    dna = {}
    window = self.window_size
    for entry in self.record:
        dna[entry.id] = str(entry.seq)
        for feature in entry.get_features('CDS'):
            orf_list = all_orf_list.get(entry.id, [])
            is_phage = int(feature.qualifiers['is_phage'][0]) if 'is_phage' in feature.qualifiers else 0
            orf_list.append(
                   {'start' : feature.start,
                    'stop'  : feature.stop,
                    'phmm'  : sum([-math.log10(x) if x != 0 else 500 for x in feature.phmm])/100,
                    'peg'   : 'peg',
                    'is_phage': is_phage  # note that we include this here, even if we are generating test data
                   }
            )
            all_orf_list[entry.id] = orf_list

    for mycontig in all_orf_list:
        orf_list = all_orf_list[mycontig]
        if not orf_list:
            # an empty list of orfs
            message("No ORFs were found in {mycontig}", "YELLOW", 'stderr')
            continue
        all_median = find_all_median(orf_list)
        lengths = []
        directions = []
        phmms = []
        for i in orf_list:
            lengths.append(abs(i['start'] - i['stop']) + 1) # find_all_median can be deleted now
            directions.append(1 if i['start'] < i['stop'] else -1)
            phmms.append(i['phmm'])
            if i['start'] < i['stop']:
                seq = dna[mycontig][i['start'] - 1 : i['stop']].upper()
            else:
                seq = reverse_complement(dna[mycontig][i['stop'] - 1 : i['start']].upper())
            my_shannon_scores.addValue(seq)

        ga_skew, gt_skew, gg_skew, gc_skew = find_avg_atgc_skew(orf_list, mycontig, dna)
        a = sum(ga_skew) / len(ga_skew)
        t = sum(gt_skew) / len(gt_skew)
        g = sum(gg_skew) / len(gg_skew)
        c = sum(gc_skew) / len(gc_skew)
        avg_at_skew, avg_gc_skew = math.fabs(a - t), math.fabs(g - c)

        for i, orf_data in enumerate(orf_list):
            # orf_length_med\tshannon_slope\tat_skew\tgc_skew\tmax_direction\tphmms
            this_orf = []  # the data for this particular orf that gets added to test_data
            # initialize
            j_start = i - int(window / 2)
            j_stop = i + int(window / 2)
            if j_start < 0:
                j_start = 0
            elif j_stop >= len(orf_list):
                j_stop = len(orf_list)
            # at and gc skews
            ja_skew = ga_skew[j_start:j_stop]
            jt_skew = gt_skew[j_start:j_stop]
            jc_skew = gc_skew[j_start:j_stop]
            jg_skew = gg_skew[j_start:j_stop]
            ja = sum(ja_skew) / len(ja_skew)
            jt = sum(jt_skew) / len(jt_skew)
            jc = sum(jc_skew) / len(jc_skew)
            jg = sum(jg_skew) / len(jg_skew)
            jat = math.fabs(ja - jt) / avg_at_skew if avg_at_skew else 0
            jgc = math.fabs(jg - jc) / avg_gc_skew if avg_gc_skew else 0
            my_length = find_median(lengths[j_start:j_stop]) - all_median
            # orf direction
            orf = []
            x = 0
            flag = 0
            for ii in directions[j_start:j_stop]:
                if ii == 1:
                    if flag == 0:
                        x += 1
                    else:
                        orf.append(x)
                        x = 1
                        flag = 0
                else:
                    if flag == 1:
                        x += 1
                    else:
                        if flag < 1 and x > 0:
                            orf.append(x)
                        x = 1
                        flag = 1
            orf.append(x)
            orf.sort()
            this_orf.append(my_length)
            if self.expand_slope:
                s = my_shannon_scores.getSlope(j_start, j_stop)
                this_orf.append(s * s)
            else:
                this_orf.append(my_shannon_scores.getSlope(j_start, j_stop))
            this_orf.append(jat)
            this_orf.append(jgc)
            this_orf.append(
                orf[len(orf) - 1] if len(orf) == 1 else orf[len(orf) - 1] + orf[len(orf) - 2]
            )
            this_orf.append(sum(phmms[j_start:j_stop]))
            assert(len(this_orf) == 6)  # confirm I added everything!
            if 'making_training_set' in self:
                this_orf.append('1' if orf_data['is_phage'] else '0')
            data.append(this_orf)
        my_shannon_scores.reset()
    return data

def make_set_train(**kwargs):
    kwargs['making_training_set'] = True
    self = Namespace(**kwargs)

    if self.output_dir:
        of = os.path.join(self.output_dir, self.make_training_data)
    else:
        of = self.make_training_data

    with open(of,'w') as outfile:
        outfile.write('orf_length_med\tshannon_slope\tat_skew\tgc_skew\tmax_direction\tphmms\tstatus\n')
        for d in measure_features(**kwargs):
            outfile.write("\t".join(map(str, d)) + "\n")



import re
import os


def is_phage_func(func):
    func = func.lower()
    func = func.replace('-', ' ')
    func = func.replace(',', ' ')
    func = func.replace('/', ' ')
    a = re.split(' ', func)
    if (
            'integrase'     in a or
            'phage'         in a or
            'lysin'         in a or
            'endolysin'     in a or
            'holin'         in a or
            'capsid'        in a or
            'tail'          in a or
            'bacteriophage' in a or
            'prophage'      in a or
            'portal'        in a or
            'terminase'     in a or
            'tapemeasure'   in a or
            'baseplate'     in a or
            'virion'        in a or
            'antirepressor' in a or
            'excisionase'   in a or
            'mobile element protein' == func or
            re.search(r"\b%s\b" % "tape measure", func) or
            re.search(r"\b%s\b" % "Cro-like repressor", func) or
            re.search(r"\b%s\b" % "CI-like repressor", func) or
            re.search(r"\b%s\b" % "rIIA lysis", func) or
            re.search(r"\b%s\b" % "rI lysis", func) or
            re.search(r"\b%s\b" % "rIIB lysis", func) or
            re.search(r"\b%s\b" % "base plate", func) or
            ("head" in a and "decoration" in a) or
            ("helix" in a and "turn" in a) or
            "HNH endonuclease" in func or
            "single-stranded DNA-binding protein" in func
    ):
        return True
    return False


def is_not_phage_func(x):
    """
    These are some annotations that are definitely not phages, but often look like them.

    :param x: the function to test
    :return: True if it is NOT a phage function.
        Note that False suggests it maybe a phage function, but may not be (we are not sure)!
    """

    x = x.lower()
    if (
        ('phage' in x) and ('shock' in x) or
        "conjugal transfer" in x or
        "conjugative" in x or
        "flagella" in x or
        "flagellar" in x or
        "flagellin" in x or
        "flagellum" in x or
        "hypothetical" in x or
        "ribosomal protein" in x or
        "secy" in x or
        "summary phrase" in x or
        "dna binding domain" in x or
        "Abortive infection bacteriophage resistance protein" in x
        ):
            return True
    return False


def is_unknown_func(x):
    x_lower = x.lower()
    if (
            (len(x) == 0) or
            # ('hypoth' in x_lower) or
            ('hypothetical' in x_lower) or
            ('conserved protein' in x_lower) or
            ('gene product' in x_lower) or
            ('interpro' in x_lower) or
            ('uncharacterized' in x_lower) or
            ('pseudogene' in x_lower) or
            ('similar to' in x_lower) or
            ('similarity' in x_lower) or
            ('glimmer' in x_lower) or
            ('unknown' in x_lower) or
            ('complete' in x_lower) or
            ('ensang' in x_lower) or
            ('unnamed' in x_lower) or
            ('expressed' in x_lower) or
            ('similar to' in x_lower) or
            (' identi' in x_lower) or
            ('ortholog of' in x_lower) or
            ('structural feature' in x_lower) or
            ('cds_' in x_lower) or
            ('predicted by psort' in x_lower) or
            ('AGR_' in x) or
            ('EG:' in x) or
            ('RIKEN' in x) or
            re.search('lmo\d+ protein', x_lower) or
            re.search('lmo\d+protein', x_lower) or
            re.search('B[sl][lr]\d', x_lower) or
            re.search('^U\d', x) or
            re.search('[a-zA-Z]{2,3}\|', x) or
            re.search('orf\d+', x_lower) or
            re.match('orf[^_]', x_lower) or
            re.match('predicted', x_lower) or
            re.match('bh\d+', x_lower) or
            re.match('y[a-z]{2,4}\\b', x) or
            re.match('[a-z]{2,3}\d+[^:\+\-0-9]', x_lower)
    ):
        return True
    return False


def downweighting_unknown_functions(self):
    for d in self.initial_tbl:
        if d[8] > 0 and is_unknown_func(d[1]):
            d[8] = 0.5
    return self.initial_tbl


#!/usr/bin/python3
__author__ = 'Przemek Decewicz'

from argparse import Namespace
from Bio import SeqIO
from glob import glob
from os import path, remove
from subprocess import call

def read_genbank(infile, infile_type, target_type):
    """
    This function reads the GenBank file and extracts protein and/or nucleotide sequences in a manner that will allow easier mapping of results against it's genome.
    """

    if infile_type == 'genbank':
        genome = { 'contigs': list(SeqIO.parse(infile, 'genbank')),
                   'proteins': [],
                   'info': {}}
    elif infile_type == 'contigs':
        genome = { 'contigs': list(SeqIO.parse(infile, 'fasta')),
                   'proteins': [],
                   'info': {}}

    if target_type == 'proteins':
        gpid_cnt = 1

        for r in genome['contigs']:
            rid = r.id
            genome['info'][rid] = {'length': len(r.seq),
                                   'description': r.description,
                                   'features': {}}

            for f in r.features:
                try:
                    genome['info'][rid]['features'][f.type] += 1
                except KeyError:
                    genome['info'][rid]['features'][f.type] = 1

                if f.type == 'CDS':
                    gpid = 'pid%07d' % gpid_cnt
                    gpid_cnt += 1

                    try:
                        pid = f.qualifiers['protein_id'][0]
                    except KeyError:
                        pid = gpid
                        f.qualifiers['protein_id'] = [gpid]

                    try:
                        product = f.qualifiers['product'][0]
                    except KeyError:
                        product = 'unknown product of protein %s' % pid

                    try:
                        aa = f.qualifiers['translation'][0]
                    except KeyError:
                        if f.location.strand == 1:
                            aa = r.seq[f.location.start:f.location.end].translate(table = 11, to_stop = True)
                        else:
                            aa = r.seq[f.location.start:f.location.end].reverse_complement().translate(table = 11, to_stop = True)

                    header = [rid, gpid, pid, str(f.location)] # product was ommited
                    genome['proteins'].append('>%s\n%s' % ('|'.join(header), aa))

        print('  Read %i contigs.' % len(genome['contigs']))
        print('  Read %i proteins.' % len(genome['proteins']))

    elif target_type == 'contigs':
        print('  Read %i contigs.' % len(genome['contigs']))

    return genome


def write_search_input(genome, target_type, outdir):
    """
    Writes protein FASTA file.
    """

    if target_type == 'proteins':
        outfile = path.join(outdir, 'prots.fasta')
        with open(outfile, 'w') as outf:
            outf.write('\n'.join(genome['proteins']))
    elif target_type == 'contigs':
        outfile = path.join(outdir, 'cont.fasta')
        with open(outfile, 'w') as outf:
            outf.write('\n'.join(genome['contigs']))

        target = path.join(outdir, 'cont_prot.fasta')
        cmd = ['transeq', '-sequence', outfile, '-outseq', target, '-clean', '-table', '11', '-frame', '6']
        print(' '.join(cmd))
        call(cmd)
        outfile = target

    return outfile


def search_target(searchfile, target_type, phmms, outdir, threads):
    """
    Performs the search of reference databases against target sequences (searchfile) using hmmsearch.
    """

    results = {'hsout': {}, # unused
               'hstbl': {}}
    print('  ----> %s' % path.basename(phmms))

    hstbl = path.join(outdir, path.basename(phmms) + '.tbl')
    hsout = path.join(outdir, path.basename(phmms) + '.out')

    cmd = ['hmmsearch', '--cpu', str(threads), '-E', '1e-10', '--domE', '1e-5', '--noali', '--tblout', hstbl, '-o', hsout, phmms, searchfile]
    # print(' '.join(cmd))
    call(cmd)

    # Read hmmsearch results
    # print('  Reading output file.')
    # results['hsout'].update(parse_hmmsearch_output(hsout, target_type))

    print('  Reading hmmsearch tbl output files.')
    results['hstbl'].update(parse_hmmersearch_tbl_output(hstbl))

    return results


def update_genbank(records, results, outfile, color):
    """
    Goes through each record and its proteins for which HMMs were assigned and adds them into annotation.
    """
    # results {sid: {pid: [hmms]}}
    for r in records:
        for f in r.features:
            if f.type != 'CDS': continue
            pid = f.qualifiers['protein_id'][0]
            try:
                phmms = results['hstbl'][r.id][pid]
            except KeyError:
                phmms = []
            for phmm in phmms:
                try:
                    f.qualifiers['phmm'].append(phmm)
                except KeyError:
                    f.qualifiers['phmm'] = [phmm]
                if color:
                    f.qualifiers['color'] = ['6']


    SeqIO.write(records, outfile, 'genbank')

    return records


def parse_hmmersearch_tbl_output(infile):
    """
    This function parses hmmsearch tbl output to further map HMM profiles identifiers into GenBank file.
    """

    results = {}    # {sid: {pid: [hmms]}}
    with open(infile) as inf:
        for line in inf:
            if line.startswith('#'): continue
            line = line.split()
            sid, gpid, pid = line[0].split('|')[:3]
            hit = f'{line[2]}:{line[4]}'
            try:
                results[sid][pid].append(hit)
            except KeyError:
                try:
                    results[sid][pid] = [hit]
                except KeyError:
                    results[sid] = {pid: [hit]}
    return results


def search_phmms(**kwargs):

    """
    Wraps phmms search.
    """

    self = Namespace(**kwargs)
    input_type = 'genbank' # alternatively contigs
    target = 'proteins' # alternatively contigs
    outfile = path.join(self.output_dir, path.basename(self.infile))


    if not self.skip_search:
        # 1 - Read GenBank file and extract required information
        print('  Reading input file: %s' % self.infile)
        genome = read_genbank(self.infile, input_type, target)


        # 2 - Prepare DNA sequence for profile searching
        print('  Writing target data.')
        searchfile = write_search_input(genome, target, self.output_dir)


        # 3 - Perform searches against prepared reference datafiles
        print('  Performing search against reference datasets.')
        results = search_target(searchfile, target, self.phmms, self.output_dir, self.threads)


        # 4 - Print the summary
        print('  Search summary.')
        for sid, pids in results['hstbl'].items():
            print(f'   - {sid}: found HMMs for {len(pids)}/{len(genome["proteins"])} proteins.')


        # 5 - Update GenBank file
        print('  Updating GenBank file.')
        genome['contigs'] = update_genbank(genome['contigs'], results, outfile, self.color)

    if not self.keep:
        if path.isfile(path.join(self.output_dir, 'prots.fasta')): remove(path.join(self.output_dir, 'prots.fasta'))
        for hmm in glob(path.join(self.output_dir, '*.hmm.*')):
            remove(hmm)

    print('  Done!')

    return outfile



#!/usr/bin/env python
import os
import sys
import gzip
from functools import reduce

from Bio import SeqIO

INSTALLATION_DIR = os.path.dirname(os.path.realpath(__file__)) + '/'
sys.path.append(INSTALLATION_DIR)

import PhiSpyModules


def main(argv):  #organismPath, output_dir, trainingFlag, INSTALLATION_DIR, evaluateOnly, threshold_for_FN, phageWindowSize, quietMode, keep):

    ######################################
    #         parse the options          #
    ######################################
    args_parser = PhiSpyModules.get_args()
    PhiSpyModules.log_and_message(f"Starting PhiSpy.py with the following arguments\n{args_parser}")

    # if we get here we need an input file
    if not args_parser.infile:
        PhiSpyModules.log_and_message("ERROR: Please provide an input file. Use -h for more options", c="RED",
                                      stderr=True, stdout=False)
        sys.exit(-1)


    ######################################
    #       add HMM search signal        #
    ######################################
    # if phmm search is required
    if args_parser.phmms:
        PhiSpyModules.log_and_message('Performing HMM search.', c="GREEN", stderr=True, stdout=False,
                                      quiet=args_parser.quiet)
        args_parser.infile = PhiSpyModules.search_phmms(**vars(args_parser))

    ######################################
    #        process input file          #
    ######################################
    # in future support other types
    # added a filter to remove short contigs. These break everything and we can't predict them to be 
    # phages anyway

    # RAE: Add support for gzipped files
    try:
        if PhiSpyModules.is_gzip_file(args_parser.infile):
            handle = gzip.open(args_parser.infile, 'rt')
        else:
            handle = open(args_parser.infile, 'r')
    except IOError as e:
        PhiSpyModules.log_and_message(f"There was an error reading {args_parser.infile}: {e}", c="RED",
                                      stderr=True, stdout=False, quiet=args_parser.quiet)
        sys.exit(20)

    args_parser.record = PhiSpyModules.SeqioFilter(filter(lambda x: len(x.seq) > args_parser.min_contig_size, SeqIO.parse(handle, "genbank")))
    handle.close()

    # do we have any records left. Yes, this bug caught me out
    ncontigs = reduce(lambda sum, element: sum + 1, args_parser.record, 0)
    if ncontigs == 0:
        msg = f"Sorry, all of the contigs in {args_parser.infile} are less than {args_parser.min_contig_size}bp.\n"
        msg += "There is no data to process"
        PhiSpyModules.log_and_message(msg, c="RED", stderr=True, stdout=False, quiet=args_parser.quiet)
        sys.exit(20)

    PhiSpyModules.log_and_message(f"Processing {ncontigs} contigs", c="GREEN", stderr=True, stdout=False,
                                  quiet=args_parser.quiet)

    ######################################
    #         make training set          #
    ######################################
    if args_parser.make_training_data:
        if not args_parser.quiet:
            PhiSpyModules.log_and_message('Making Training Set...', c="GREEN", stderr=True, stdout=False,
                                          quiet=args_parser.quiet)
        my_make_train_flag = PhiSpyModules.make_set_train(**vars(args_parser))
        exit()

    ######################################
    #         make testing set           #
    ######################################
    PhiSpyModules.log_and_message('Making Testing Set...', c="GREEN", stderr=True, stdout=False,
                                  quiet=args_parser.quiet)
    args_parser.test_data = PhiSpyModules.measure_features(**vars(args_parser))

    ######################################
    #         do classification          #
    ######################################
    PhiSpyModules.log_and_message('Start Classification Algorithm...', c="GREEN", stderr=True, stdout=False,
                                  quiet=args_parser.quiet)
    args_parser.rfdata = PhiSpyModules.call_randomforest(**vars(args_parser))
    args_parser.initial_tbl = PhiSpyModules.make_initial_tbl(**vars(args_parser))

    ######################################
    #         i dont know what           #
    ######################################
    ###### added in this version 2.2 #####
    if (args_parser.training_set == 'data/trainSet_genericAll.txt'):
        PhiSpyModules.log_and_message('As the training flag is zero, down-weighting unknown functions', c="RED",
                                      stderr=True, stdout=False, quiet=args_parser.quiet)
        args_parser.initial_tbl = PhiSpyModules.downweighting_unknown_functions(args_parser)

    ######################################
    #         do evaluation              #
    ######################################
    PhiSpyModules.log_and_message('Evaluating...', c="GREEN", stderr=True, stdout=False, quiet=args_parser.quiet)
    PhiSpyModules.fixing_start_end(**vars(args_parser))
    # don't forget to close the log!
    PhiSpyModules.log_and_message(f'Done!!! Output is in {args_parser.output_dir} and the log is in {args_parser.log}',
                                  c="WHITE", stderr=True, stdout=False, quiet=args_parser.quiet)


if __name__== "__main__":
    main(sys.argv)




"""
A module to write the output in different formats
"""

import os
import gzip


from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation
from collections import OrderedDict

import PhiSpyModules.version as version
from .formatting import message
from .helper_functions import is_gzip_file
import logging

__author__ = 'Rob Edwards'

def log_and_message(msg, c="WHITE", stderr=False, stdout=False, quiet=False, loglevel="INFO"):
    """
    Write the message to both the log and an output stream. By default we will just log the message in the
    log.

    Note that we also adhere to the quiet option of self, and will only write to the log
    Set either stderr or stdout to true to write to those streams too (but you will need to reset quiet if appropriate)

    :param self: The arg parser object
    :param msg: the message to write
    :param c: the color to write the message
    :param stderr: write the message to stderr
    :param stdout: write the message to stdout
    :param loglevel: the logging level. See https://docs.python.org/3/library/logging.html#levels for a list
    :return:
    """

    if loglevel == 'CRITICAL':
        logging.getLogger('PhiSpy').critical(msg.strip())
    elif loglevel == 'ERROR':
        logging.getLogger('PhiSpy').error(msg.strip())
    elif loglevel == 'WARNING':
        logging.getLogger('PhiSpy').warning(msg.strip())
    elif loglevel == 'DEBUG':
        logging.getLogger('PhiSpy').debug(msg.strip())
    else:
        logging.getLogger('PhiSpy').info(msg.strip())

    if not quiet:
        if stderr:
            message(msg,c, "stderr")
        if stdout:
            message(msg, c, "stdout")


def write_gff3(self, pp):
    """
    Write GFF3 code. This was adapted from code contribued by [Jose Francisco Sanchez-Herrero]
    (https://github.com/JFsanchezherrero/)

    :param self: the data object
    :param pp: array of pp dictionaries
    :return: None
    """

    log_and_message("Writing GFF3 output file", c="GREEN", stderr=True, quiet=self.quiet)
    # GFF output
    out_gff = open(os.path.join(self.output_dir, self.file_prefix + "prophage.gff3"), "w")
    out_gff.write("##gff-version 3")
    out_gff.write("\n")
    
    # loop through pp results
    for i in pp:
        # GFF
        # strand is not known...
        out_gff.write(str(pp[i]['contig']) +
                      '\tPhiSpy\tprophage_region\t' +
                      str(pp[i]['start']) + '\t' + 
                      str(pp[i]['stop']) + '\t.\t.' + 
                      ' \t.\tID=pp' + str(i))
        out_gff.write('\n')

        if 'att' not in pp[i]:
            pp[i]['att'] = ""
        else:
            # attL
            out_gff.write(str(pp[i]['contig']) +
                          '\tPhiSpy\tattL\t' +
                          str(pp[i]['att'][0]) + '\t' + 
                          str(pp[i]['att'][1]) + '\t.\t.' + 
                          ' \t.\tID=pp' + str(i))
            out_gff.write('\n')

            # attR
            out_gff.write(str(pp[i]['contig']) +
                          '\tPhiSpy\tattR\t' +
                          str(pp[i]['att'][2]) + '\t' +
                          str(pp[i]['att'][3]) + '\t.\t.' +
                          ' \t.\tID=pp' + str(i))
            out_gff.write('\n')
        
    out_gff.close()


def write_genbank(self, pp):
    """
    Write prophages and their potential attachment sites in updated input GenBank file.
    :param self: the data object
    :param pp: array of pp dictionaries
    :return: None
    """

    log_and_message("Writing GenBank output file", c="GREEN", stderr=True, quiet=self.quiet)
    prophage_feature_type = 'misc_feature'  # / prophage_region
    outfile = os.path.join(self.output_dir, self.file_prefix + os.path.basename(self.infile))
    for i in pp:
        self.record.get_entry(pp[i]['contig']).append_feature(SeqFeature(
                    location=FeatureLocation(pp[i]['start'], pp[i]['stop']),
                    type=prophage_feature_type,
                    strand=1,
                    qualifiers=OrderedDict(
                        {'note': f'prophage region pp{i} identified with PhiSpy v{version.__version__}'}
                    )))
        if 'atts' in pp[i]:
            self.record.get_entry(pp[i]['contig']).append_feature(SeqFeature(
                        location=FeatureLocation(int(pp[i]['att'][0]), int(pp[i]['att'][1])) +
                                 FeatureLocation(int(pp[i]['att'][2]), int(pp[i]['att'][3])),
                        type='repeat_region',
                        strand=1,
                        qualifiers=OrderedDict({'note': f'prophage region pp{i} potential attachment sites'})))

    # are we writing a gzip file
    if is_gzip_file(self.infile):
        handle = gzip.open(outfile, 'wt')
    else:
        handle = open(outfile, 'w')

    SeqIO.write(self.record, handle, 'genbank')


def write_phage_and_bact(self, pp, dna):
    """
    Separate out the phage and bacterial fractions into fasta files
    :param self: the data object
    :param pp: array of pp dictionaries
    :param dna: the DNA sequence object
    :return: None
    """
    log_and_message('Writing bacterial and phage DNA as fasta', c="GREEN", stderr=True, quiet=self.quiet)
    phage_out = open(os.path.join(self.output_dir, self.file_prefix + "phage.fasta"), "w")
    phage_genbank = open(os.path.join(self.output_dir, self.file_prefix + "phage.gbk"), "w")
    bacteria_out = open(os.path.join(self.output_dir, self.file_prefix + "bacteria.fasta"), "w")
    bacteria_genbank = open(os.path.join(self.output_dir, self.file_prefix + "bacteria.gbk"), "w")
    
    contig_to_phage = {}
    for i in pp:
        contig = pp[i]['contig']  # contig name
        if contig not in contig_to_phage:
            contig_to_phage[contig] = set()
        contig_to_phage[contig].add(i)

    for contig in dna:
        if contig not in contig_to_phage:
            bacteria_out.write(f">{contig}\n{dna[contig]}\n")
            SeqIO.write(self.record.get_entry(contig), bacteria_genbank, "genbank")
            continue
        pps1 = list(contig_to_phage[contig])
        pps = sorted(pps1, key=lambda k: pp[k]['start'])
        bactstart = 0
        dnaseq = ""
        hostcounter = 0
        for ppnum in pps:
            pphagestart = pp[ppnum]['start']
            pphagestop = pp[ppnum]['stop']

            phageseq = dna[contig][pphagestart:pphagestop]
            phage_out.write(f">{contig}_{pphagestart}_{pphagestop} [pp {ppnum}]\n{phageseq}\n")
            pp_gbk = self.record.get_entry(contig)[pphagestart:pphagestop]
            # set the locus tag
            pp_gbk.name += f"_PP{ppnum}"
            # set the Accession
            pp_gbk.id += f"_PP{ppnum}"
            #set the definition line
            pp_gbk.description += f" prophage PP{ppnum} on {contig} from {pphagestart} to {pphagestop}"
            SeqIO.write(pp_gbk, phage_genbank, "genbank")

            dnaseq += dna[contig][bactstart:pphagestart - 1]
            dnaseq += "N" * len(phageseq)
            host_gbk = self.record.get_entry(contig)[bactstart:pphagestart - 1]
            hostcounter += 1
            host_gbk.name += f"_region_{hostcounter}"
            host_gbk.id += f"_region_{hostcounter}"
            host_gbk.description += f" region {hostcounter} on {contig} from {bactstart} to {pphagestart - 1}"
            SeqIO.write(host_gbk, bacteria_genbank, "genbank")
            bactstart = pphagestop + 1

        dnaseq += dna[contig][bactstart:]
        host_gbk = self.record.get_entry(contig)[bactstart:]
        hostcounter += 1
        host_gbk.name += f"_region_{hostcounter}"
        host_gbk.id += f"_region_{hostcounter}"
        host_gbk.description += f" region {hostcounter} on {contig} from {bactstart} onwards"
        bacteria_out.write(f">{contig} [phage regions replaced with N]\n{dnaseq}\n")
        SeqIO.write(host_gbk, bacteria_genbank, "genbank")

    phage_out.close()
    phage_genbank.close()
    bacteria_out.close()
    bacteria_genbank.close()


def write_prophage_coordinates(self, pp):
    """
    Write the coordinates and other details about the prophages
    :param self: the data object
    :param pp: array of pp dictionaries
    :return: None
    """

    log_and_message("Writing prophage_coordinates output file", c="GREEN", stderr=True, quiet=self.quiet)
    with open(os.path.join(self.output_dir, self.file_prefix + 'prophage_coordinates.tsv'), 'w') as out:
        for i in pp:
            if 'atts' not in pp[i]:
                pp[i]['atts'] = ""
            locs = [
                "pp" + str(i),
                pp[i]['contig'],
                pp[i]['start'],
                pp[i]['stop'],
                pp[i]['atts']
            ]
            out.write("\t".join(map(str, locs)) + "\n")


def write_prophage_tbl(self, pp):
    """
    Create a prophage_tbl file from our pp dictionary
    :param self: the data object
    :param pp: array of pp dictionaries
    :return: None
    """

    log_and_message("Writing prophage.tbl output file", c="GREEN", stderr=True, quiet=self.quiet)
    with open(os.path.join(self.output_dir, self.file_prefix + "prophage.tbl"), 'w') as out:
        for i in pp:
            locs = [
                "pp_" + str(i),
                pp[i]['contig'],
                pp[i]['start'],
                pp[i]['stop']
            ]
            out.write(locs[0] + "\t" + "_".join(map(str, locs[1:])) + "\n")


def write_prophage_tsv(self, pp):
    """
    Create a tsv with headers for this data. Issue #28 item 2
    :param self: the data object
    :param pp: array of pp dictionaries
    :return: None
    """

    log_and_message("Writing prophage.tsv output file", c="GREEN", stderr=True, quiet=self.quiet)
    with open(os.path.join(self.output_dir, self.file_prefix + "prophage.tsv"), 'w') as out:
        out.write("Prophage number\tContig\tStart\tStop\n")
        for i in pp:
            locs = [
                "pp_" + str(i),
                pp[i]['contig'],
                pp[i]['start'],
                pp[i]['stop']
            ]
            out.write("\t".join(map(str, locs)) + "\n")


def write_test_data(self, pp):
    """
    Write the testing measurements
    :param self: the data object
    :param measurements: the measurements
    :return: None
    """
    log_and_message("Writing test_data output file", c="GREEN", stderr=True, quiet=self.quiet)
    with open(os.path.join(self.output_dir, self.file_prefix + "test_data.tsv"), 'w') as out:
        if self.phmms:
            out.write('identifier\torf_length_med\tshannon_slope\tat_skew\tgc_skew\tmax_direction\tphmms\tstatus\n')
        else:
            out.write('identifier\torf_length_med\tshannon_slope\tat_skew\tgc_skew\tmax_direction\tstatus\n')

        for i, d in enumerate(self.test_data):
            out.write("\t".join(map(str, [self.initial_tbl[i][0]] + d)) + "\n")


def prophage_measurements_to_tbl(inputf, outputf):
    try:
        f = open(inputf, 'r')
        fw = open(outputf, 'w')
    except IOError as e:
        log_and_message(f"There was an error trying to convert the measurements to a tbl: {e}", c="RED", stderr=True)
        return
    pp = {}
    ppindx = 0
    prev_contig = None
    inphage = False
    header = f.readline()
    for line in f:
        temp = line.strip().split("\t")
        if int(temp[9]) > 0:
            newphage = False
            if temp[2] != prev_contig:
                newphage = True
            if not inphage:
                newphage = True
            if newphage:
                ppindx += 1
                pp[ppindx] = {}
                pp[ppindx]['contig'] = temp[2]
                pp[ppindx]['start'] = min(int(temp[3]), int(temp[4]))
                pp[ppindx]['stop'] = max(int(temp[3]), int(temp[4]))
            else:
                pp[ppindx]['stop'] = max(int(temp[3]), int(temp[4]))
            inphage = True
            prev_contig = temp[2]
        else:
            inphage = False
    for i in pp:
        fw.write("pp_" + str(i) + "\t" + pp[i]['contig'] + "_" + str(pp[i]['start']) + "_" + str(pp[i]['stop']) + "\n")
    f.close()
    fw.close()





"""
This is a helper module to just make colored text if we are *not* redirected

Colors come from https://stackoverflow.com/questions/287871/print-in-terminal-with-colors
Testing redirect comes from
https://stackoverflow.com/questions/1512457/determining-if-stdout-for-a-python-process-is-redirected
"""

import os
import sys

from .errors import ColorNotFoundError


"""
if os.fstat(0) == os.fstat(1):
    we are not redirected
"""

"""
Colors that you can import and make the text look pretty


"""

__author__ = 'Rob Edwards'


class Colors(object):
    color = {
        'HEADER': '\033[95m',
        'OKBLUE': '\033[94m',
        'OKGREEN': '\033[92m',
        'WARNING': '\033[93m',
        'FAIL': '\033[91m',
        'ENDC': '\033[0m',
        'BOLD': '\033[1m',
        'UNDERLINE': '\033[4m',
        'PINK': '\033[95m',
        'BLUE': '\033[94m',
        'GREEN': '\033[92m',
        'YELLOW': '\033[93m',
        'RED': '\033[91m',
        'WHITE': '\033[0m',
        }


def message(msg, c, stream):
    """
    Print a message to stderr using color
    :param msg: the message to print
    :param c: the colour to use
    :param stream: either stderr or stdout
    :return: nothing
    """

    c = c.upper()
    # we strip any newline off message in case it was added and then we add it here! Ensures always only one newline
    msg = msg.strip()
    if c not in Colors.color:
        raise ColorNotFoundError(f"Color {c} was not found")

    if stream.lower() == 'stderr':
        if os.fstat(0) == os.fstat(1):
            #  stderr is not redirected
            sys.stderr.write(f"{Colors.color[c]}{msg}{Colors.color['ENDC']}\n")
        else:
            sys.stderr.write(f"{msg}\n")
    elif stream.lower() == 'stdout':
        if os.fstat(0) == os.fstat(1):
            #  stderr is not redirected
            sys.stdout.write(f"{Colors.color[c]}{msg}{Colors.color['ENDC']}\n")
        else:
            sys.stdout.write(f"{msg}\n")
    else:
        raise IOError(f"There is no IO stream {stream}")



import setuptools
from distutils.core import Extension
import os

# Read the markdown files for the long description
def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        if os.path.exists(filename):
            with open(filename, "r", encoding=encoding) as f:
                buf.append(f.read())
    return sep.join(buf)

long_description = read('README.md', 'CHANGES.md')

def get_version():
    with open("VERSION", 'r') as f:
        v = f.readline().strip()
        return v

def main():
    setuptools.setup(
        name="PhiSpy",
        version=get_version(),
        description="Prophage finder using multiple metrics",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Rob Edwards",
        platforms='any',
        keywords="phage prophage bioinformatics microbiology bacteria genome genomics",
        author_email="raedwards@gmail.com",
        url='https://github.com/linsalrob/PhiSpy',
        license='The MIT License (MIT)',
        scripts=['PhiSpy.py'],
        packages=setuptools.find_packages(),
        ext_modules=[Extension("PhiSpyRepeatFinder", sources=["src/repeatFinder.cpp"], language='c++')],
        include_package_data=True,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Intended Audience :: Science/Research',
            'License :: OSI Approved :: MIT License',
            'Natural Language :: English',
            'Operating System :: MacOS :: MacOS X',
            'Operating System :: POSIX :: Linux',
            'Operating System :: Unix',
            'Programming Language :: Python :: 3.0',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
        ],
        install_requires = [
            'biopython>=1.74',
            'numpy>=1.17.0',
            'scikit-learn>=0.21.3',
        ]
        )

if __name__ == "__main__":
    main()



# read and set the version

from .seqio_filter import SeqioFilter
from .makeTest import make_set_train, measure_features
from .classification import call_randomforest, make_initial_tbl
from .protein_functions import downweighting_unknown_functions, is_phage_func, is_unknown_func
from .evaluation import fixing_start_end
from .helper_functions import get_args, print_list, is_gzip_file
from .pathtype import PathType
from .writers import write_gff3, write_genbank, write_phage_and_bact, write_prophage_tbl, write_prophage_tsv
from .writers import prophage_measurements_to_tbl
from .writers import log_and_message
from .search_phmms import search_phmms
from .formatting import message
from .errors import ColorNotFoundError, NoBasesCounted

from .version import __version__

__all__ = ['SeqioFilter',
           'make_set_train', 'measure_features',
           'call_randomforest', 'make_initial_tbl',
           'downweighting_unknown_functions', 'is_phage_func', 'is_unknown_func',
           'fixing_start_end',
           'get_args', 'print_list', 'is_gzip_file',
           'PathType',
           'write_gff3', 'write_genbank', 'write_phage_and_bact', 'write_prophage_tbl', 'write_prophage_tsv',
           'prophage_measurements_to_tbl', 'log_and_message',
           'search_phmms',
           'message',
            'ColorNotFoundError', 'NoBasesCounted'
           ]



import re
import sys
import os
import pkg_resources
from io import TextIOWrapper
from sklearn.cluster import KMeans
from sklearn.ensemble import RandomForestClassifier
import numpy as np
from argparse import Namespace

from .protein_functions import is_phage_func, is_unknown_func, is_not_phage_func
from .writers import log_and_message

def find_training_genome(training_flag):
    try:
        f = pkg_resources.resource_stream('PhiSpyModules', 'data/trainingGenome_list.txt')
    except IOError as e:
        log_and_message(f"There was an error opening data/trainingGenome_list.txt: {e}\n", c="RED", stderr=True,
                        loglevel="CRITICAL")
        return ''

    for line in f:
        temp = re.split('\t', line.decode().strip())
        if int(temp[0]) == training_flag:
            f.close()
            return temp[1].strip()
    return ''


def call_randomforest(**kwargs):
    training_file = kwargs['training_set']
    test_data = kwargs['test_data']

    if not pkg_resources.resource_exists('PhiSpyModules', training_file):
        log_and_message(f"FATAL: Can not find data file {training_file}\n", c="RED", stderr=True, loglevel="CRITICAL")
        sys.exit(-1)
    strm = pkg_resources.resource_stream('PhiSpyModules', training_file)
    log_and_message(f"Using training set in {training_file}")
    train_data = np.genfromtxt(TextIOWrapper(strm), delimiter="\t", skip_header=1, filling_values=1)

    if 'phmms' not in kwargs:
        train_data = np.delete(train_data, 5, 1)
        test_data = np.delete(train_data, 5, 1)
    """
    Przemek's comment
    by default 10 until version 0.22 where default is 100
    number of estimators also implies the precision of probabilities, generally 1/n_estimators
    in R's randomForest it's 500 and the usage note regarding number of trees to grow says:
    "This should not be set to too small a number, to ensure that every input row gets predicted at least a few times."
    """
    log_and_message(f"Running the random forest classifier with {kwargs['randomforest_trees']} trees and {kwargs['threads']} threads")
    clf = RandomForestClassifier(n_estimators=kwargs['randomforest_trees'], n_jobs=kwargs['threads'])
    clf.fit(train_data[:, :-1], train_data[:, -1].astype('int'))
    return clf.predict_proba(test_data)[:,1]


def my_sort(orf_list):
    n = len(orf_list)
    i = 1
    while i <= n:
        j = i + 1
        while j < n:
            flag = 0
            # direction for both
            if orf_list[i]['start'] < orf_list[i]['stop']:
                dir_i = 1
            else:
                dir_i = -1
            if orf_list[j]['start'] < orf_list[j]['stop']:
                dir_j = 1
            else:
                dir_j = -1

            # check whether swap need or not
            if dir_i == dir_j:
                if orf_list[i]['start'] > orf_list[j]['start']:
                    flag = 1
            else:
                if dir_i == 1:
                    if orf_list[i]['start'] > orf_list[j]['stop']:
                        flag = 1
                else:
                    if orf_list[i]['stop'] > orf_list[j]['start']:
                        flag = 1
            # swap
            if flag == 1:
                temp = orf_list[i]
                orf_list[i] = orf_list[j]
                orf_list[j] = temp
            j = j+1
        i = i+1
    return orf_list


def find_mean(all_len):
    s = 0.0
    for i in all_len:
        s = s + i
    return float(s)/len(all_len)


def calc_pp(func):
    func = func.replace('-', ' ')
    func = func.replace(',', ' ')
    func = func.replace('/', ' ')
    func = func.lower()
    x = 0
    if is_phage_func(func):
        x = 1
    elif is_unknown_func(func):
        x = 0.5
    elif is_not_phage_func(func):
        x = 0

    # a few special cases
    if 'recombinase' in func or 'integrase' in func:
        x = 1.5

    return x


def make_initial_tbl(**kwargs):
    self = Namespace(**kwargs)
    data = []
    x = []
    for entry in self.record:
        for feature in entry.get_features('CDS'):
            ft = {
                'fig': feature.id,
                'function': feature.function,
                'contig': entry.id,
                'start': feature.start,
                'stop': feature.stop,
                'rank': 0.0,
                'status': 0,
                'pp': calc_pp(feature.function),
            }
            x.append(ft)

    ranks = [[] for n in range(len(x))]
    for j, val in enumerate(self.rfdata):
        for k in range(j-int(self.window_size/2), j+int(self.window_size/2)):
            if k < 0 or k >= len(x) or j >= len(x) or x[k]['contig'] != x[j]['contig']:
                continue
            ranks[k].append(val)

    # calculate threshold
    y = []
    j = 0
    while j < len(x):
        x[j]['rank'] = sum(ranks[j]) / len(ranks[j])
        x[j]['extra'] = ranks[j]
        y.append(x[j]['rank'])
        j = j+1

    y2 = np.array(y).reshape(-1, 1)
    km = KMeans(n_clusters=2)
    km.fit(y2)
    centers = km.cluster_centers_
    threshold = max(centers[0][0], centers[1][0])
    log_and_message(f"Rank threshold calculated as {threshold}")
    """
    Note added by Rob:
    At this point we have the classifications for each ORF and we want to take a sliding window and decide where 
    the phage should start. We have two calculations for a threshold for the rank: either the kmeans centers and 
    finding things above the larger center or just a plain threshold.
    
    """

    """
    Eventually each row has:
        0. gene id           6. rank                12. start of attR
        1. function          7. my status           13. end of attR
        2. contig            8. pp                  14. sequence of attL
        3. start             9. final status        15. sequence of attR
        4. stop             10. start of attL       16. Reason for att site choice
        5. position         11. end of attL
        
    However, at this point we only have  0 .. 8
    """


    for i in range(len(x)):
        status = 1 if x[i]['rank'] > threshold else 0
        thisrow = [
            x[i]['fig'],
            x[i]['function'],
            x[i]['contig'],
            x[i]['start'],
            x[i]['stop'],
            i,
            x[i]['rank'],
            status,
            x[i]['pp']
        ]
        data.append(thisrow)
    return data


